const five = require('johnny-five');
const { Notification } = require('electron')
const SerialPort = require('serialport')


const TEXTS = {
    ERROR: 'Error',
    PERFECT: 'Perfecto!',
    READY: 'Echidna está conectado y listo para ser controlado con Scratch!',
    NOT_CONNECTED: 'No encuentro al echidna, asegurate de que está conectado, sal de la aplicación y vuelve a arrancarla'
}

const LEDS = {
    red: 'd13',
    yellow: 'd12',
    green: 'd11'
};

const RGB = {
    red: 'd9',
    green: 'd5',
    blue: 'd6'
};

const DIGITAL_PINS = {
    d2: '2',
    d3: '3',
    d4: '4',
    d5: '5',
    d6: '6',
    d7: '7',
    d8: '8',
    d9: '9',
    d11: '11',
    d12: '12',
    d13: '13',
}

const ANALOG_PINS = {
    a0: 'A0',
    a1: 'A1',
    a2: 'A2',
    a3: 'A3',
    a4: 'A4',
    a5: 'A5',
    a6: 'A6',
    a7: 'A7'
}

function fail(e) {
    let m = (e.message == 'No connected device found') ?
        TEXTS.NOT_CONNECTED : e
    let n = new Notification({
        title: TEXTS.ERROR,
        body: m
    });
    n.show();
}

function ready() {
    // Aquí this es un objeto Board
    console.log('Arduino is ready.');
    let n = new Notification({
        title: TEXTS.PERFECT,
        body: TEXTS.READY
    });
    n.show();
}

function resetEchidna(board) {
    console.log("reseting echidna ...");
    for (let pin in DIGITAL_PINS) {
        if (board.pins[pin] &&
            (board.pins[pin].mode == five.Pin.OUTPUT || board.pins[pin].mode == five.Pin.PWM)) {
            board.pins[pin].write(0);
            board.pins[pin] = undefined;
        }
    }
}

// esta función crea los pines si no existen o los vuelve a crear
// en el modo que se solicita si el modo era otro.
function configurePin(pinName, mode, board, io = null) {
    if (board.pins == undefined) {
        board.pins = {}
    }

    let pin;
    if (mode == five.Pin.ANALOG) {
        pin = ANALOG_PINS[pinName];
    } else {
        pin = DIGITAL_PINS[pinName];
    }

    if (board.pins[pinName] == undefined || board.pins[pinName].mode != mode) {
        console.log("redefino pin " + pin);
        board.pins[pinName] = new five.Pin({
            pin: pin,
            mode: mode
        });
        board.pins[pinName].on('data', value => {
            // if (pinName == 'd2' || pinName == 'd3') {
            //     console.log({ id: pinName, value: value });
            // }
            if (mode == five.Pin.INPUT) {
                io.emit('digitalInputValue', { id: pinName, value: value });
            }
            if (mode == five.Pin.ANALOG) {
                io.emit('analogInputValue', { id: pinName, value: value });
            }
        })

    }
}

function led(data, board) {
    console.log(data)

    let pin = LEDS[data.name];
    console.log(pin);
    //if(board.digitalOutput[pin] == undefined) return;

    configurePin(pin, five.Pin.OUTPUT, board);

    if (data.action == 'on') {
        board.pins[pin].write(1);
        //board.digitalWrite(pin, 1);
    } else if (data.action == 'off') {
        board.pins[pin].write(0);
        //board.digitalWrite(pin, 0);
    } else {
        console.log(data.action);
        console.log("orden no reconocida");
    }
}

function rgb(data, board) {
    console.log(data);

    for (color in RGB) {
        let pin = RGB[color];
        configurePin(pin, five.Pin.PWM, board);
    }

    if (data.action == 'color') {
        // No podemos usar la función write para escribir un tren de pulsos
        // PWM, tenemos que usar la función analogWrite del objeto board
        // board.pins[RGB["red"]].write(data.color["red"]);
        // board.pins[RGB["green"]].write(data.color["green"]);
        // board.pins[RGB["blue"]].write(data.color["blue"]);
        console.log(DIGITAL_PINS[RGB["red"]]);
        board.analogWrite(DIGITAL_PINS[RGB["red"]], data.color["red"]);
        board.analogWrite(DIGITAL_PINS[RGB["green"]], data.color["green"]);
        board.analogWrite(DIGITAL_PINS[RGB["blue"]], data.color["blue"]);
    } else {
        console.log(data.action);
        console.log("orden no reconocida");
    }
}

function servo(data, board) {
    console.log(data);

    // No sé lo que pasa, pero en el momento en que se crean los
    // servos, el pin 9 PWM (rgb rojo) deja de funcionar.
    // Así que en el momento en que se use el servo, el pin 9 se jode
    if (!board.servos) board.servos = {};
    board.servos[data['pin']] = new five.Servo(DIGITAL_PINS[data['pin']]);

    let pin = data['pin'];
    if (pin != 'd4' && pin != 'd7' && pin != 'd8') {
        console.log("pin no válido para servo")
        return;
    }
    let angle = data["angle"]
    board.servos[pin].to(angle);
}

function piezo(data, board) {
    console.log(data);
    if (data["action"] == "frequency") {
        board.piezo.frequency(data["frequency"], data["duration"]);
    }

    if (data["action"] == "play") {
        board.piezo.play(data["tune"]);
    }
}

function digitalOutput(data, board) {
    console.log(data);

    configurePin(data.id, five.Pin.OUTPUT, board);

    board.pins[data.id].write(data.value);
}

function digitalInput(data, board, io) {
    console.log(data);

    configurePin(data.id, five.Pin.INPUT, board, io);

    /* board.pins[data.id].query(function(state) {
        io.emit('digitalInputValue', { id: data.id, value: state.value })
    }) */
}

function analogInput(data, board, io) {
    console.log(data);
    configurePin(data.id, five.Pin.ANALOG, board, io);
}

function define_events(io, board) {
    return function(client) {
        // client es un socket
        console.log("Cliente conectado con id: " + client.id);

        io.clients((error, clients) => {
            if (error) throw error;
            //console.log(clients); // => [PZDoMHjiu8PYfRiKAAAF, Anw2LatarvGVVXEIAAAD]
        });

        client.on('join', function(handshake) {
            console.log(handshake);
        });

        client.on('client_disconnect', data => {
            resetEchidna(board, io);
            console.log("Cliente desconectado ..." + data.id);
        })

        client.on('reset', data => {
            resetEchidna(board, io);
        });

        // Outputs
        client.on('led', data => {
            led(data, board);
        });

        client.on('rgb', data => {
            rgb(data, board);
        });

        client.on('servo', data => {
            servo(data, board);
        });

        client.on('piezo', data => {
            piezo(data, board);
        });

        client.on('digitalOutput', data => {
            digitalOutput(data, board);
        })

        client.on('digitalInput', data => {
            digitalInput(data, board, io);
        })

        client.on('analogInput', data => {
            analogInput(data, board, io);
        })
    }
}

function EchidnaBoard(io) {
    this.board = new five.Board({ repl: false });
    this.board.on("fail", fail.bind(this));
    this.board.on("ready", ready);

    io.on("connection", define_events(io, this.board));
    io.on("disconnect", function() {
        console.log('Disconnection ...');
        console.log(client);
    });


}

exports.EchidnaBoard = { EchidnaBoard }